# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import functools
import json
import logging

from crunchyroll.apis import ApiInterface
from crunchyroll.apis.android import AndroidApi
from crunchyroll.apis.ajax import AjaxApi
from crunchyroll.apis.scraper import ScraperApi
from crunchyroll.constants import META, AJAX, ANDROID
from crunchyroll.apis.errors import *
from crunchyroll.models import *
from crunchyroll.util import return_collection

logger = logging.getLogger('crunchyroll.apis.meta')

def require_session_started(func):
    """Check if API sessions are started and start them if not
    """
    @functools.wraps(func)
    def inner_func(self, *pargs, **kwargs):
        if not self.session_started:
            logger.info('Starting session for required meta method')
            self.start_session()
        return func(self, *pargs, **kwargs)
    return inner_func

def require_android_logged_in(func):
    """Check if andoid API is logged in and login if not, implies
    `require_session_started`
    """
    @functools.wraps(func)
    @require_session_started
    def inner_func(self, *pargs, **kwargs):
        if not self._android_api.logged_in:
            logger.info('Logging into android API for required meta method')
            if not self.has_credentials:
                raise ApiLoginFailure(
                    'Login is required but no credentials were provided')
            self._android_api.login(account=self._state['username'],
                password=self._state['password'])
        return func(self, *pargs, **kwargs)
    return inner_func

def optional_android_logged_in(func):
    """Check if andoid API is logged in and login if credentials were provided,
    implies `require_session_started`
    """
    @functools.wraps(func)
    @require_session_started
    def inner_func(self, *pargs, **kwargs):
        if not self._android_api.logged_in and self.has_credentials:
            logger.info('Logging into android API for optional meta method')
            self._android_api.login(account=self._state['username'],
                password=self._state['password'])
        return func(self, *pargs, **kwargs)
    return inner_func

def require_ajax_logged_in(func):
    """Check if ajax API is logged in and login if not
    """
    @functools.wraps(func)
    def inner_func(self, *pargs, **kwargs):
        if not self._ajax_api.logged_in:
            logger.info('Logging into AJAX API for required meta method')
            if not self.has_credentials:
                raise ApiLoginFailure(
                    'Login is required but no credentials were provided')
            self._ajax_api.User_Login(name=self._state['username'],
                password=self._state['password'])
        return func(self, *pargs, **kwargs)
    return inner_func

def optional_ajax_logged_in(func):
    """Check if ajax API is logged in and login if credentials available
    """
    @functools.wraps(func)
    def inner_func(self, *pargs, **kwargs):
        if not self._ajax_api.logged_in and self.has_credentials:
            logger.info('Logging into AJAX API for optional meta method')
            self._ajax_api.User_Login(name=self._state['username'],
                password=self._state['password'])
        return func(self, *pargs, **kwargs)
    return inner_func

class MetaApi(ApiInterface):
    """High level interface to crunchyroll
    """

    def __init__(self, username=None, password=None, state=None):
        self._state = {
            'username': username,
            'password': password,
        }
        self._ajax_api = AjaxApi()
        self._android_api = AndroidApi()
        if state is not None:
            self.set_state(state)

    @property
    def session_started(self):
        return self._ajax_api.session_started and \
            self._android_api.session_started

    @property
    def logged_in(self):
        return self._ajax_api.logged_in and self._android_api.logged_in

    @property
    def has_credentials(self):
        return self._state['username'] is not None and \
            self._state['password'] is not None

    def get_state(self):
        return json.dumps({
            'meta':     self._state,
            'ajax':     self._ajax_api.get_state(),
            'android':  self._android_api.get_state(),
        })

    def set_state(self, state):
        # TODO: error handling here
        decoded_state = json.loads(state)
        self._state = decoded_state['meta']
        self._ajax_api.set_state(decoded_state['ajax'])
        self._android_api.set_state(decoded_state['android'])

    @optional_android_logged_in
    def is_premium(self, media_type):
        """Get if the user is premium for a given media type

        @param str media_type       should be one of ANDROID.MEDIA_TYPE_*
        @return bool
        """
        return self._android_api.is_premium(media_type)

    def start_session(self):
        """Start the underlying APIs sessions

        Calling this is not required, it will be called automatically if
        a method that needs a session is called

        @return bool
        """
        self._android_api.start_session()
        return self.session_started

    @require_session_started
    def login(self, username, password):
        """Login with the given username/email and password

        Calling this method is not required if credentials were provided in
        the constructor, but it could be used to switch users or something maybe

        @return bool
        """
        # we could get stuck in an inconsistent state if got an exception while
        # trying to login with different credentials than what is stored so
        # we rollback the state to prevent that
        state_snapshot = self._state.copy()
        try:
            self._ajax_api.User_Login(name=username, password=password)
            self._android_api.login(account=username, password=password)
        except Exception as err:
            # something went wrong, rollback
            self._state = state_snapshot
            raise err
        self._state['username'] = username
        self._state['password'] = password
        return self.logged_in

    @optional_android_logged_in
    @return_collection(Series)
    def list_anime_series(self, sort=META.SORT_ALPHA, limit=META.MAX_SERIES, offset=0):
        """Get a list of anime series

        @param str sort     pick how results should be sorted, should be one
                                of META.SORT_*
        @param int limit    limit number of series to return, there doesn't
                                seem to be an upper bound
        @param int offset   list series starting from this offset, for pagination
        @return list<crunchyroll.models.Series>
        """
        result = self._android_api.list_series(
            media_type=ANDROID.MEDIA_TYPE_ANIME,
            filter=sort,
            limit=limit,
            offset=offset)
        return result

    @optional_android_logged_in
    @return_collection(Series)
    def list_drama_series(self, sort=META.SORT_ALPHA, limit=META.MAX_SERIES, offset=0):
        """Get a list of drama series

        @param str sort     pick how results should be sorted, should be one
                                of META.SORT_*
        @param int limit    limit number of series to return, there doesn't
                                seem to be an upper bound
        @param int offset   list series starting from this offset, for pagination
        @return list<crunchyroll.models.Series>
        """
        result = self._android_api.list_series(
            media_type=ANDROID.MEDIA_TYPE_DRAMA,
            filter=sort,
            limit=limit,
            offset=offset)
        return result

    @optional_android_logged_in
    @return_collection(Series)
    def search_anime_series(self, query_string):
        """Search anime series list by series name, case-sensitive

        @param str query_string     string to search for, note that the search
                                        is very simplistic and only matches against
                                        the start of the series name, ex) search
                                        for "space" matches "Space Brothers" but
                                        wouldn't match "Brothers Space"
        @return list<crunchyroll.models.Series>
        """
        result = self._android_api.list_series(
            media_type=ANDROID.MEDIA_TYPE_ANIME,
            filter=ANDROID.FILTER_PREFIX + query_string)
        return result

    @optional_android_logged_in
    @return_collection(Series)
    def search_drama_series(self, query_string):
        """Search drama series list by series name, case-sensitive

        @param str query_string     string to search for, note that the search
                                        is very simplistic and only matches against
                                        the start of the series name, ex) search
                                        for "space" matches "Space Brothers" but
                                        wouldn't match "Brothers Space"
        @return list<crunchyroll.models.Series>
        """
        result = self._android_api.list_series(
            media_type=ANDROID.MEDIA_TYPE_DRAMA,
            filter=ANDROID.FILTER_PREFIX + query_string)
        return result

    @optional_android_logged_in
    @return_collection(Media)
    def list_media(self, series, sort=META.SORT_DESC, limit=META.MAX_MEDIA, offset=0):
        """List media for a given series or collection

        @param crunchyroll.models.Series series the series to search for
        @param str sort                         choose the ordering of the
                                                    results, only META.SORT_DESC
                                                    is known to work
        @param int limit                        limit size of results
        @param int offset                       start results from this index,
                                                    for pagination
        @return list<crunchyroll.models.Media>
        """
        params = {
            'sort': sort,
            'offset': offset,
            'limit': limit,
        }
        params.update(self._get_series_query_dict(series))
        result = self._android_api.list_media(**params)
        return result

    @optional_android_logged_in
    @return_collection(Media)
    def search_media(self, series, query_string):
        """Search for media from a series starting with query_string, case-sensitive

        @param crunchyroll.models.Series series     the series to search in
        @param str query_string                     the search query, same restrictions
                                                        as `search_anime_series`
        @return list<crunchyroll.models.Media>
        """
        params = {
            'sort': ANDROID.FILTER_PREFIX + query_string,
        }
        params.update(self._get_series_query_dict(series))
        result = self._android_api.list_media(**params)
        return result

    @optional_ajax_logged_in
    def get_media_stream(self, media_item, format, quality):
        """Get the stream data for a given media item

        @param crunchyroll.models.Media media_item
        @param int format
        @param int quality
        @return crunchyroll.models.MediaStream
        """
        result = self._ajax_api.VideoPlayer_GetStandardConfig(
            media_id=media_item.media_id,
            video_format=format,
            video_quality=quality)
        return MediaStream(result)

    @optional_ajax_logged_in
    def get_stream_info(self, media_item, format, quality):
        result = self._ajax_api.VideoEncode_GetStreamInfo(
            media_id=media_item.media_id,
            video_format=format,
            video_encode_quality=quality)
        return StreamInfo(result)

    @return_collection(SubtitleStub)
    def get_subtitle_stubs(self, media_item):
        result = self._ajax_api.Subtitle_GetListing(media_id=media_item.media_id)
        return XmlModel(result)['subtitle']

    def unfold_subtitle_stub(self, subtitle_stub):
        """Turn a SubtitleStub into a full Subtitle object

        @param crunchyroll.models.SubtitleStub subtitle_stub
        @return crunchyroll.models.Subtitle
        """
        return Subtitle(self._ajax_api.Subtitle_GetXml(
            subtitle_script_id=int(subtitle_stub.id)))

    @optional_ajax_logged_in
    def get_stream_formats(self, media_item):
        """Get the available media formats for a given media item

        @param crunchyroll.models.Media
        @return dict
        """
        scraper = ScraperApi(self._ajax_api._connector)
        formats = scraper.get_media_formats(media_item.media_id)
        return formats

    @require_android_logged_in
    @return_collection(Series)
    def list_queue(self, media_types=[META.TYPE_ANIME, META.TYPE_DRAMA]):
        """List the series in the queue, optionally filtering by type of media

        @param list<str> media_types    a list of media types to filter the queue
                                            with, should be of META.TYPE_*
        @return list<crunchyroll.models.Series>
        """
        result = self._android_api.queue(media_types='|'.join(media_types))
        return [queue_item['series'] for queue_item in result]

    @require_android_logged_in
    def add_to_queue(self, series):
        """Add a series to the queue

        @param crunchyroll.models.Series series
        @return bool
        """
        result = self._android_api.add_to_queue(series_id=series.series_id)
        return result

    @require_android_logged_in
    def remove_from_queue(self, series):
        """Remove a series from the queue

        @param crunchyroll.models.Series series
        @return bool
        """
        result = self._android_api.remove_from_queue(series_id=series.series_id)
        return result

    def _get_series_query_dict(self, series):
        """Pick between collection_id and series_id params in series models for the
        Android API

        @param crunchyroll.models.Series series
        @return dict
        """
        if hasattr(series, 'series_id'):
            return {'series_id': series.series_id}
        else:
            return {'collection_id': series.collection_id}



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import logging

import requests

from crunchyroll.apis import ApiInterface
from crunchyroll.constants import AJAX
from crunchyroll.apis.errors import *

logger = logging.getLogger('crunchyroll.apis.ajax')

def make_ajax_api_method(req_method, secure=False):
    def outer_func(func):
        def inner_func(self, **kwargs):
            kwargs['req'] = 'RpcApi' + func.__name__
            kwargs['current_page'] = AJAX.API_CURRENT_PAGE
            req_url = self._build_request_url(secure)
            req_func = self._build_request(req_method, req_url, secure, params=kwargs)
            try:
                response = func(self, req_func)
                self._last_response = response
            except ApiLoginFailure as err:
                raise err
            except Exception as err: # TODO: make this more specific
                logger.warn('Caught exception of class: %s', err.__class__.__name__)
                raise ApiNetworkException(err)
            if not (response.ok and response.headers['Content-Type'] == 'text/xml'):
                raise ApiBadResponseException(response)
            else:
                return response.content
        return inner_func
    return outer_func

class AjaxApi(ApiInterface):
    """AJAX call API
    """

    METHOD_POST = 'POST'
    METHOD_GET  = 'GET'

    def __init__(self, state=None):
        self._connector = requests.Session()
        self._last_response = None
        if state is not None:
            self.set_state(state)

    def _build_request_url(self, secure):
        proto = AJAX.PROTOCOL_SECURE if secure else AJAX.PROTOCOL_INSECURE
        return AJAX.API_URL.format(protocol=proto)

    def _build_request(self, req_method, req_url, secure, params):
        def req_func():
            logger.debug('Sending %s request to "%s" with params: %r',
                req_method, req_url, params)
            try:
                func = getattr(self._connector, req_method.lower())
            except AttributeError:
                raise ApiException('Invalid request method')
            try:
                if secure and req_method == self.METHOD_POST:
                    # wouldn't make sense to send data on a GET request
                    resp = func(req_url, data=params)
                else:
                    resp = func(req_url, params=params)
            except requests.RequestException as err:
                raise ApiNetworkException(err)
            logger.debug('Received response code: %d', resp.status_code)
            return resp
        return req_func

    @property
    def session_started(self):
        return True

    @property
    def logged_in(self):
        return AJAX.COOKIE_USERID in self._connector.cookies

    def get_state(self):
        state_string = json.dumps(dict(self._connector.cookies))
        logger.debug('Generated state: %s', state_string)
        return state_string

    def set_state(self, state):
        logger.debug('Loading state: %s', state)
        cookie_jar = json.loads(state)
        self._connector.cookies.update(cookie_jar)

    @make_ajax_api_method(METHOD_POST, True)
    def User_Login(self, req_func):
        """
        @param str name     username or email address
        @param str password user's password
        @param str fail_url (optional) URL to redirect to after login
        """
        response = req_func()
        if not self.logged_in:
            raise ApiLoginFailure(response)
        return response

    @make_ajax_api_method(METHOD_POST)
    def Subtitle_GetXml(self, req_func):
        """Get just the <subtitle/> tag info

        @param int subtitle_script_id
        """
        return req_func()

    @make_ajax_api_method(METHOD_POST)
    def Subtitle_GetListing(self, req_func):
        """
        @param int media_id
        """
        return req_func()

    @make_ajax_api_method(METHOD_POST)
    def VideoEncode_GetStreamInfo(self, req_func):
        """
        @param int video_encode_quality
        @param int video_format
        @param int media_id
        """
        return req_func()

    @make_ajax_api_method(METHOD_POST)
    def VideoPlayer_GetStandardConfig(self, req_func):
        """
        @param int media_id
        @param int video_format
        @param int video_quality
        @param int auto_play        (optional)
        @param str aff              (optional)
        @param int show_pop_out_controls    (optional)
        @param str pop_out_disable_message  (optional)
        """
        return req_func()

    @make_ajax_api_method(METHOD_POST)
    def VideoPlayer_GetChromelessConfig(self, req_func):
        """
        @param int media_id
        """
        return req_func()

    @make_ajax_api_method(METHOD_POST)
    def VideoPlayer_GetMediaMetadata(self, req_func):
        """
        @param int media_id
        @param str video_format
        @param str video_encode_quality
        """
        return req_func()

    @make_ajax_api_method(METHOD_GET)
    def VideoPlayer_GetAutoAdvanceTarget(self, req_func):
        """
        @param int media_id
        """
        return req_func()

    @make_ajax_api_method(METHOD_POST)
    def Media_GetRecommendedMedia(self, req_func):
        """
        @param int media_id
        """
        return req_func()



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import functools
import pipes
import xml.etree.ElementTree as ET
try:
    from HTMLParser import HTMLParser
except ImportError:
    from html.parser import HTMLParser

logger = logging.getLogger('crunchyroll.util')

# use a singleton parser
_html_parser = HTMLParser()

def html_unescape(html_string):
    return _html_parser.unescape(html_string)

def return_collection(collection_type):
    """Change method return value from raw API output to collection of models
    """
    def outer_func(func):
        @functools.wraps(func)
        def inner_func(self, *pargs, **kwargs):
            result = func(self, *pargs, **kwargs)
            return map(collection_type, result)
        return inner_func
    return outer_func

def parse_xml_string(xml_string):
    return ET.fromstring(xml_string)

def xml_node_to_string(xml_node):
    return ET.tostring(xml_node)

def format_rtmpdump_args(rtmp_data):
    arg_string = '-r {url} -W {swf_url} -T {token} -y {file} ' \
        '-p {page_url} -t {url}'
    return arg_string.format(**dict([(k, pipes.quote(v)) for (k,v) in rtmp_data.items()]))

# NullHandler was added in py2.7
if hasattr(logging, 'NullHandler'):
    NullHandler = logging.NullHandler
else:
    class NullHandler(logging.Handler):
        def handle(self, record):
            pass

        def emit(self, record):
            pass

        def createLock(self):
            self.lock = None

LOG_FORMAT = logging.Formatter('[%(asctime)s] %(levelname)8s - %(name)s: %(message)s')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Crunchyroll Library
-------------------

A library to interface with Crunchyroll's various APIs and utilites to work
with the returned data

Set environment var CRUNCHYROLL_DEBUG to '1' to enable debug logging to
stderr

:copyright: (c) 2013 by Alex Headley
:license: GNU General Public License v2+, see LICENCE for more details
"""

__title__           = 'crunchyroll'
__version__         = '1.0.1'
__author__          = 'Alex Headley'
__author_email__    = 'aheadley@waysaboutstuff.com'
__license__         = 'GNU General Public License v2+'
__copyright__       = 'Copyright 2013 Alex Headley'
__url__             = 'https://github.com/aheadley/python-crunchyroll'
__description__     = """
A library to interface with Crunchyroll's various APIs and utilites to work
with the returned data
""".strip()

# set default logging handler
import os
import logging

from crunchyroll.util import NullHandler, LOG_FORMAT

logger = logging.getLogger(__title__)

if os.environ.get('CRUNCHYROLL_DEBUG', False):
    _log_handler = logging.StreamHandler()
    _log_handler.setFormatter(LOG_FORMAT)
    _log_level = logging.DEBUG
else:
    _log_handler = NullHandler()
    _log_level = logging.WARN

logger.setLevel(_log_level)
logger.addHandler(_log_handler)

logger.debug('%s module init finished', __title__)



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import locale
import json
import logging

import requests

from crunchyroll.apis import ApiInterface
from crunchyroll.constants import ANDROID
from crunchyroll.apis.errors import *

logger = logging.getLogger('crunchyroll.apis.android')

def make_android_api_method(req_method, secure=True, version=0):
    """Turn an AndroidApi's method into a function that builds the request,
    sends it, then passes the response to the actual method. Should be used
    as a decorator.
    """
    def outer_func(func):
        def inner_func(self, **kwargs):
            req_url = self._build_request_url(secure, func.__name__, version)
            req_func = self._build_request(req_method, req_url, params=kwargs)
            response = req_func()
            func(self, response)
            return response
        return inner_func
    return outer_func

# TODO: should rename this to MobileApi or AppApi, I doubt the iPhone API is
# different (is there even an iPhone app?)
class AndroidApi(ApiInterface):
    """Android (and probably iPhone) API interface

    Optional API method parameters are marked as such, or have a default
    value specified.

    Only hard-subbed and SD streams are supported, HD/soft-subs are simply
    not available through the API.
    @link http://www.crunchyroll.com/forumtopic-777139/are-streams-hd-on-the-new-kindle-fire-hd?fpid=41205823
    """

    METHOD_GET      = 'GET'
    METHOD_POST     = 'POST'

    def __init__(self, state=None):
        """Init object, optionally with previously stored session and/or auth
        tokens
        """
        self._connector = requests.Session()
        self._request_headers = {
            'X-Android-Device-Manufacturer':
                ANDROID.DEVICE_MANUFACTURER,
            'X-Android-Device-Model':
                ANDROID.DEVICE_MODEL,
            'X-Android-Device-Product':
                ANDROID.DEVICE_PRODUCT,
            # changing this to '1' doesn't seem to have an effect
            'X-Android-Device-Is-GoogleTV': '1',
            'X-Android-SDK': ANDROID.SDK_VERSION,
            'X-Android-Release': ANDROID.RELEASE_VERSION,
            'X-Android-Application-Version-Code': ANDROID.APP_CODE,
            'X-Android-Application-Version-Name': ANDROID.APP_PACKAGE,
            'User-Agent': ANDROID.USER_AGENT,
        }
        self._state_params = {
            'session_id': None,
            'auth': None,
            'user': None,
        }
        self._user_data = None
        # for debugging
        self._last_response = None
        # dunno what these are for yet, seems to be a way to tell the client
        # to do something
        self._session_ops = []
        if state is not None:
            self.set_state(state)
        logger.info('Initialized state: %r', self._state_params)

    def _get_locale(self):
        """Get the current locale with dashes (-) and underscores (_) removed

        Ex: en-US -> enUS
        """
        return locale.getdefaultlocale()[0].replace('_', '').replace('-', '')

    def _get_base_params(self):
        """Get the params that will be included with every request
        """
        base_params = {
            'locale':       self._get_locale(),
            'device_id':    ANDROID.DEVICE_ID,
            'device_type':  ANDROID.APP_PACKAGE,
            'access_token': ANDROID.ACCESS_TOKEN,
            'version':      ANDROID.APP_CODE,
        }
        base_params.update(dict((k, v) \
            for k, v in self._state_params.iteritems() \
                if v is not None))
        return base_params

    def _do_post_request_tasks(self, response_data):
        """Handle actions that need to be done with every response

        I'm not sure what these session_ops are actually used for yet, seems to
        be a way to tell the client to do *something* if needed.
        """
        try:
            sess_ops = response_data.get('ops', [])
        except AttributeError:
            pass
        else:
            self._session_ops.extend(sess_ops)

    def _build_request(self, method, url, params=None):
        """Build a function to do an API request

        "We have to go deeper" or "It's functions all the way down!"
        """
        full_params = self._get_base_params()
        if params is not None:
            full_params.update(params)
        try:
            request_func = lambda u, d: \
                getattr(self._connector, method.lower())(u, params=d,
                    headers=self._request_headers)
        except AttributeError:
            raise ApiException('Invalid request method')
        # TODO: need to catch a network here and raise as ApiNetworkException

        def do_request():
            logger.debug('Sending %s request "%s" with params: %r',
                method, url, full_params)
            try:
                resp = request_func(url, full_params)
                logger.debug('Received response code: %d', resp.status_code)
            except requests.RequestException as err:
                raise ApiNetworkException(err)
            try:
                resp_json = resp.json()
            except TypeError:
                resp_json = resp.json
            try:
                is_error = resp_json['error']
            except TypeError:
                raise ApiBadResponseException(resp.content)
            if is_error:
                raise ApiError('%s: %s' % (resp_json['code'], resp_json['message']))
            else:
                self._last_response = resp
                data = resp_json['data']
                self._do_post_request_tasks(data)
                return data
        return do_request

    def _build_request_url(self, secure, api_method, version):
        """Build a URL for a API method request
        """
        if secure:
            proto = ANDROID.PROTOCOL_SECURE
        else:
            proto = ANDROID.PROTOCOL_INSECURE
        req_url = ANDROID.API_URL.format(
            protocol=proto,
            api_method=api_method,
            version=version
        )
        return req_url

    @property
    def session_started(self):
        return self._state_params.get('session_id', None) is not None

    @property
    def logged_in(self):
        return self._state_params['auth'] is not None

    def is_premium(self, media_type):
        """Get if the session is premium for a given media type

        @param str media_type       Should be one of ANDROID.MEDIA_TYPE_*
        @return bool
        """
        if self.logged_in:
            if media_type in self._user_data['premium']:
                return True
        return False

    def get_state(self):
        state = {
            'state_params': self._state_params,
            'cookies': dict(self._connector.cookies),
            'user_data': self._user_data,
        }
        return json.dumps(state)

    def set_state(self, state):
        loaded_state = json.loads(state)
        self._state_params.update(loaded_state['state_params'])
        self._connector.cookies.update(loaded_state['cookies'])
        self._user_data = loaded_state['user_data']

    @make_android_api_method(METHOD_POST, False)
    def start_session(self, response):
        """
        This is the only method that doesn't go over HTTPS (for some reason). Must
        be called before anything else or you get an "unauthorized request" error.

        I don't know what the duration param does in any of these methods

        @param int duration (optional)
        """
        self._state_params['session_id'] = response['session_id']
        self._state_params['country_code'] = response['country_code']

    @make_android_api_method(METHOD_POST)
    def end_session(self, response):
        """
        Should probably be called after ``logout``
        """
        self._state_params['session_id'] = None

    @make_android_api_method(METHOD_POST)
    def login(self, response):
        """
        Login using email/username and password, used to get the auth token

        @param str account
        @param str password
        @param int duration (optional)
        """
        self._state_params['auth'] = response['auth']
        self._user_data = response['user']
        if not self.logged_in:
            raise ApiLoginFailure(response)

    @make_android_api_method(METHOD_POST)
    def logout(self, response):
        """
        Auth param is not actually required, will be included with requests
        automatically after logging in
        """
        self._state_params['auth'] = None
        self._user_data = None

    @make_android_api_method(METHOD_POST)
    def authenticate(self, response):
        """
        This does not appear to be used, might refresh auth token though.

        @param str auth
        @param int duration (optional)
        """
        pass

    @make_android_api_method(METHOD_GET)
    def list_series(self, response):
        """
        Get the list of series, default limit seems to be 20.

        @param str media_type   one of ANDROID.MEDIA_TYPE_*
        @param str filter       one of ANDROID.FILTER_*, (optional)
        @param int offset=0     pick the index to start at, is not multiplied
                                    by limit or anything
        @param int limit=20     does not seem to have an upper bound
        @param str fields       comma separated list of ANDROID.FIELD.* for extra
                                    info, this must be used to get things like
                                    video links which aren't in the default
                                    field set (optional)
        """
        pass

    @make_android_api_method(METHOD_GET)
    def list_media(self, response):
        """
        Get the list of videos for a series

        Only collection_id *or* series_id should be given, the API will
        probably give an error if either both or none are given

        @param int collection_id
        @param int series_id
        @param str sort=ANDROID.FILTER_POPULAR
        @param int offset=0
        @param int limit=20
        @param str fields       comma separated list of ANDROID.FIELD.* for extra
                                    info, this must be used to get things like
                                    video links which aren't in the default
                                    field set (optional)
        """
        pass

    @make_android_api_method(METHOD_GET)
    def info(self, response):
        """
        Get info about a specific video

        Same deal as `list_media`, only pass collection_id or series_id, not
        both or none

        @param int media_id
        @param int collection_id
        @param int series_id
        @param str fields       comma separated list of ANDROID.FIELD.* for extra
                                    info, this must be used to get things like
                                    video links which aren't in the default
                                    field set (optional)
        """
        pass

    """ ** METHODS PAST THIS POINT ARE UNTESTED ** """

    @make_android_api_method(METHOD_POST)
    def add_to_queue(self, response):
        """
        @param int series_id
        """
        pass

    @make_android_api_method(METHOD_GET)
    def categories(self, response):
        """
        @param str media_type   probably should be one of ANDROID.MEDIA_TYPE_*
        """
        pass

    @make_android_api_method(METHOD_GET)
    def queue(self, response):
        """
        @param str media_types  | (pipe) separated list of ANDROID.MEDIA_TYPE_*
        """
        pass

    @make_android_api_method(METHOD_GET)
    def recently_watched(self, response):
        """
        @param str media_types
        @param int offset
        @param int limit
        """
        pass

    @make_android_api_method(METHOD_POST)
    def remove_from_queue(self, response):
        """
        @param int series_id
        """
        pass

    @make_android_api_method(METHOD_POST)
    def signup(self, response):
        """
        @param str email
        @param str password
        @param str username
        @param str first_name
        @param str last_name
        @param int duration
        """
        pass

    @make_android_api_method(METHOD_POST)
    def free_trial_start(self, response):
        """
        @param str sku
        @param str currency_code
        @param str first_name
        @param str last_name
        @param str cc
        @param str exp_month
        @param str exp_year
        @param str zip
        @param str address_1
        @param str address_2
        @param str city
        @param str state
        @param str country_code
        """
        pass

    @make_android_api_method(METHOD_POST)
    def forgot_password(self, response):
        """
        @param str email
        """
        pass

    @make_android_api_method(METHOD_GET)
    def free_trial_info(self, response):
        """
        """
        pass

    @make_android_api_method(METHOD_GET)
    def list_ads(self, response):
        """
        @param int media_id
        """
        pass

    @make_android_api_method(METHOD_POST)
    def log_ad_requested(self, response):
        """
        @param str ad_network
        """
        pass

    @make_android_api_method(METHOD_POST)
    def log_ad_served(self, response):
        """
        @param str ad_network
        """
        pass

    @make_android_api_method(METHOD_POST)
    def log_first_launch(self, response):
        """
        """
        pass

    @make_android_api_method(METHOD_POST)
    def log_impression(self, response):
        """
        @param str view
        """
        pass

    @make_android_api_method(METHOD_POST)
    def log_install_referrer(self, response):
        """
        @param str referrer
        """
        pass

    @make_android_api_method(METHOD_POST)
    def log(self, response):
        """
        @param str event
        @param int media_id
        @param int playhead
        @param int elapsed
        @param int elapsed_delta
        @param int video_id
        @param int video_encode_id
        """
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import re
import logging
import functools

from crunchyroll.util import parse_xml_string, return_collection, xml_node_to_string
from crunchyroll.subtitles import SubtitleDecrypter, SRTFormatter, ASS4plusFormatter
from crunchyroll.constants import META

logger = logging.getLogger('crunchyroll.models')

class DictModel(object):
    def __init__(self, data):
        if not isinstance(data, dict):
            raise TypeError('DictModel can only be initialized with a dict')
        else:
            self._data = data

    def __getattr__(self, name):
        try:
            item = self._data.get(name)
        except KeyError as err:
            raise AttributeError(err)
        try:
            return DictModel(item)
        except TypeError:
            return item

    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self._data))

class XmlModel(object):
    def __init__(self, node):
        if isinstance(node, basestring):
            try:
                node = parse_xml_string(node)
            except Exception as err:
                raise ValueError(err)
        elif isinstance(node, XmlModel):
            logger.debug('Creating new %s with node=%r', self.__class__.__name__,
                node)
            node = node._data
        elif node is None:
            raise ValueError('XmlModel node cannot be NoneType')
        self._data = node

    def __getattr__(self, name):
        try:
            return self._data.attrib.get(name)
        except KeyError as err:
            raise AttributeError(err)

    def __repr__(self):
        name = self.__class__.__name__
        try:
            return '<%s(id=%s)>' % (name, self.id)
        except (AttributeError, TypeError):
            return '<%s(%s)>' % (name, self.tag_name)

    def __str__(self):
        return self.text if self.text is not None else repr(self)

    __unicode__ = __str__

    def __getitem__(self, name):
        return map(XmlModel, self.findall('./' + name))

    @property
    def text(self):
        return self._data.text

    @property
    def tag_name(self):
        return self._data.tag

    def findall(self, query):
        return map(XmlModel, self._data.findall(query))

    def findfirst(self, query):
        try:
            return self.findall(query)[0]
        except IndexError:
            return None

class Series(DictModel):
    pass

class Media(DictModel):
    pass

class SubtitleStub(XmlModel):
    LANG_UNKNOWN    = 'UNKNOWN'

    @property
    def language(self):
        lang = re.search(r'^\[.*\]\s*(.*)', self.title)
        if lang:
            lang_string = lang.group(1)
        else:
            lang_string = self.LANG_UNKNOWN
        logger.debug('%r language: %r -> %r', self, self.title, lang_string)
        return lang_string

    @property
    def is_default(self):
        return self.default == '1'

class Subtitle(XmlModel):
    def __init__(self, node):
        super(Subtitle, self).__init__(node)
        self._decrypter = SubtitleDecrypter()

    def decrypt(self):
        return StyledSubtitle(self._decrypter.decrypt(
            self._decrypter._build_encryption_key(int(self.id)),
            self['iv'][0].text.decode('base64'),
            self['data'][0].text.decode('base64')))

class StyledSubtitle(XmlModel):
    def get_ass_formatted(self):
        formatter = ASS4plusFormatter()
        return formatter.format(self)

    def get_srt_formatted(self):
        formatter = SRTFormatter()
        return formatter.format(self)

def require_not_upsell(func):
    @functools.wraps(func)
    def inner_func(self, *pargs, **kwargs):
        if not self.is_upsell:
            return func(self, *pargs, **kwargs)
        else:
            return None

class StreamInfo(XmlModel):
    @property
    def is_upsell(self):
        return bool(self['upsell'])

    @property
    def rtmp_data(self):
        data = {
            'url':         self.findfirst(
                './/host').text,
            'file':         self.findfirst(
                './/file').text,
            'token':        self.findfirst(
                './/token').text,
            'swf_url':      META.SWF_URL,
            'page_url':     META.PAGE_URL,
        }
        return data

    @property
    def duration(self):
        return float(self.findfirst(
            './/metadata/duration').text)

    @property
    def resolution(self):
        width = self.findfirst(
            '//metadata/width').text
        height = self.findfirst(
            '//metadata/height').text
        return (int(width), int(height))

class MediaStream(XmlModel):
    @property
    def stream_info(self):
        return StreamInfo(self._data.findall('.//{default}preload/stream_info')[0])

    @property
    def default_subtitles(self):
        return Subtitle(self.findfirst('.//{default}preload/subtitle'))

    @property
    @return_collection(SubtitleStub)
    def subtitle_stubs(self):
        return self.findall('.//{default}preload/subtitles/subtitle')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import unittest
import os

from crunchyroll.apis.meta import MetaApi
from crunchyroll.apis.errors import *

CRUNCHYROLL_USERNAME = os.environ.get('CRUNCHYROLL_USERNAME')
CRUNCHYROLL_PASSWORD = os.environ.get('CRUNCHYROLL_PASSWORD')

CREDS_AVAILABLE = CRUNCHYROLL_USERNAME and CRUNCHYROLL_PASSWORD

skip_if_no_creds = unittest.skipUnless(CREDS_AVAILABLE, 'Login credentials not available')

class TestIntegration(unittest.TestCase):
    def setUp(self):
        self.api = MetaApi()

    @skip_if_no_creds
    def test_login(self):
        self.api.login(username=CRUNCHYROLL_USERNAME, password=CRUNCHYROLL_PASSWORD)

    def test_login_fail(self):
        with self.assertRaises(ApiLoginFailure):
            self.api.login(username='example-bad-username', password='example-bad-password')

    def test_search(self):
        self._series = self.api.search_anime_series('Space Brothers')[0]
        self.assertEqual('Space Brothers', self._series.name)

    def test_list_media(self):
        self._media = self.api.list_media(self._series)
        self.assertEqual(99, len([ep for ep in self._media if not ep.clip]))
        self.assertEqual('Life Changes, Promises Don\'t')

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging

logger = logging.getLogger('crunchyroll.apis')

# urllib3 gives annoying warnings when used with older python versions that don't
# matter to us, so we disable them
# @source https://urllib3.readthedocs.org/en/latest/security.html#insecureplatformwarning
try:
    import requests
    requests.packages.urllib3.disable_warnings()
    logger.info('Disabled urllib3 warnings')
except Exception as err:
    logger.info('Couldnt disable urllib3 warnings: %s', err)

class ApiInterface(object):
    """This will be the basis for the shared API interfaces once the Ajax and
    Web APIs have been implemented
    """

    @property
    def session_started(self):
        """Check if the API session has started
        """
        raise NotImplemented

    @property
    def logged_in(self):
        """Check if a user has been logged in through the API
        """
        raise NotImplemented

    def get_state(self):
        """Get the internal state so that the session can be
        resumed without logging in again between runs
        """
        raise NotImplemented

    def set_state(self, state):
        """Set the internal state to resume without logging in
        again
        """
        raise NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

class API:
    BASE_DOMAIN = 'crunchyroll.com'
    PROTOCOL_INSECURE   = 'http'
    PROTOCOL_SECURE     = 'https'

    LOG_FORMAT          = '%(asctime)s - %(levelname)s - [%(name)s] %(message)s'

class ANDROID(API):
    ACCESS_TOKEN                = 'Scwg9PRRZ19iVwD'
    API_DOMAIN                  = 'api.' + API.BASE_DOMAIN
    API_URL                     = '{protocol}://' + API_DOMAIN + '/' \
                                    '{api_method}.{version}.json'

    # eventually these should be used to identify this library so CR can get
    # usage statistics about people using third-party apps or something, maybe
    DEVICE_MANUFACTURER = 'unknown'
    DEVICE_MODEL        = 'google_sdk'
    DEVICE_PRODUCT      = 'google_sdk'
    USER_AGENT          = 'Dalvik/1.6.0 (Linux; U; Android 4.2; google_sdk Build/JB_MR1)'
    # this should probably be replaced with a UUIDv4 in api init
    DEVICE_ID           = '00000000-18c4-ade8-ffff-ffff99d603a9'
    SDK_VERSION         = '17'
    RELEASE_VERSION     = '4.2'
    APP_CODE            = '66'
    APP_VERSION_NAME    = '0.7.9'
    APP_PACKAGE         = 'com.crunchyroll.crunchyroid'

    # dunno what these are for yet
    PREMIUM_FLAG_ANIME      = 'anime'
    PREMIUM_FLAG_DRAMA      = 'drama'
    PREMIUM_FLAG_TAISENG    = 'taiseng'

    # media type flags
    MEDIA_TYPE_ANIME    = 'anime'
    MEDIA_TYPE_DRAMA    = 'drama'

    # filter flags
    FILTER_POPULAR      = 'popular'
    FILTER_SIMULCAST    = 'simulcast'
    FILTER_UPDATED      = 'updated'
    FILTER_ALPHA        = 'alpha'
    FILTER_NEWEST       = 'newest'
    FILTER_PREFIX       = 'prefix:' # search maybe?
    FILTER_TAG          = 'tag:' # need to find a list of tags
    # media listing must use one of these
    FILTER_ASC          = 'asc'
    FILTER_DESC         = 'desc'

    class FIELD:
        # these are apparently like database fields or something that you can add
        # to the list_series, list_media, and info methods to get additional or only
        # specific info. of particular note is the MEDIA_STREAM_DATA field
        # to get the streams
        GENERAL_MOST_LIKELY_MEDIA   = 'most_likely_media'
        GENERAL_ORDERING            = 'ordering'
        MEDIA_ID                    = 'media.media_id'
        MEDIA_NAME                  = 'media.name'
        MEDIA_EPISODE_NUMBER        = 'media.episode_number'
        MEDIA_SCREENSHOT_IMAGE      = 'media.screenshot_image'
        MEDIA_STREAM_DATA           = 'media.stream_data'
        MEDIA_FREE_AVAILABLE        = 'media.free_available'
        MEDIA_PREMIUM_AVAILABLE     = 'media.premium_available'
        MEDIA_AVAILABILITY_NOTES    = 'media.availablity_notes'
        MEDIA_PLAYHEAD              = 'media.playhead'
        MEDIA_TYPE                  = 'media.media_type'
        MEDIA_DESCRIPTION           = 'media.description'
        SERIES                      = 'series'
        SERIES_ID                   = 'series.series_id'
        SERIES_NAME                 = 'series.name'
        SERIES_PUBLISHER_NAME       = 'series.publisher_name'
        SERIES_YEAR                 = 'series.year'
        SERIES_SCREENSHOT_IMAGE     = 'series.screenshot_image'
        SERIES_LANDSCAPE_IMAGE      = 'series.landscape_image'
        SERIES_PORTRAIT_IMAGE       = 'series.portrait_image'
        SERIES_MEDIA_COUNT          = 'series.media_count'
        SERIES_MEDIA_TYPE           = 'series.media_type'
        SERIES_IN_QUEUE             = 'series.in_queue'
        SERIES_DESCRIPTION          = 'series.description'
        IMAGE_FWIDE_URL             = 'image.fwide_url'
        IMAGE_WIDESTAR_URL          = 'image.widestar_url'
        IMAGE_LARGE_URL             = 'image.large_url'
        IMAGE_FULL_URL              = 'image.full_url'

class AJAX(API):
    API_DOMAIN                  = 'www.' + API.BASE_DOMAIN
    API_URL                     = '{protocol}://' + API_DOMAIN + '/xml/'
    API_CURRENT_PAGE            = API.PROTOCOL_INSECURE + '://' + API_DOMAIN + '/'

    HEADER_REFERRER     = 'http://static.ak.crunchyroll.com/flash/' \
        '20130201144858.bd8118f7c58d1da788d88782497e30a4/StandardVideoPlayer.swf'
    HEADER_ORIGIN       = 'http://static.ak.crunchyroll.com'
    HEADER_USER_AGENT   = ANDROID.USER_AGENT

    COOKIE_USERID       = 'c_userid'
    COOKIE_USERKEY      = 'c_userkey'
    COOKIE_VISITOR      = 'c_visitor'
    COOKIE_SESSID       = 'sess_id'

    class VIDEO:
        FORMATS             = [
            '1080p',
            '720p',
            '480p',
            '360p',
        ]

class SCRAPER(API):
    API_DOMAIN                  = 'www.' + API.BASE_DOMAIN
    API_URL                     = '{protocol}://' + API_DOMAIN + '/'

    class VIDEO(AJAX.VIDEO):
        FORMAT_PARAMS       = dict((f, f[-1]+f[:-1]) for f in AJAX.VIDEO.FORMATS)
        FORMAT_PATTERN      = r'swfobject\.embedSWF.*express_install\.swf\',' \
            '.*video_format%3D(\d+).*video_quality%3D(\d+)%26.*,loadBrandedPlayerAd.*'



class META(API):
    MAX_SERIES          = 500
    MAX_MEDIA           = 1000

    TYPE_ANIME          = ANDROID.MEDIA_TYPE_ANIME
    TYPE_DRAMA          = ANDROID.MEDIA_TYPE_DRAMA

    # sorting for series lists
    SORT_POPULAR        = ANDROID.FILTER_POPULAR
    SORT_SIMULCAST      = ANDROID.FILTER_SIMULCAST
    SORT_UPDATED        = ANDROID.FILTER_UPDATED
    SORT_ALPHA          = ANDROID.FILTER_ALPHA
    SORT_NEWEST         = ANDROID.FILTER_NEWEST
    # sorting for media lists
    SORT_ASC            = ANDROID.FILTER_ASC
    SORT_DESC           = ANDROID.FILTER_DESC

    VIDEO = AJAX.VIDEO

    # normally you would need the player revision to generate this URL, but
    # fortunately CR provides us with the latest version to the CPA.swf without
    # the revision in the path
    SWF_URL             = 'http://static.ak.crunchyroll.com/swf/ChromelessPlayerApp.swf'
    PAGE_URL            = AJAX.API_CURRENT_PAGE



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

class ApiException(Exception):
    """Base class for exceptions thrown by the API classes
    """
    pass

class ApiNetworkException(ApiException):
    """We couldn't talk to the API because the internet tubes are clogged or
    something
    """
    pass

class ApiBadResponseException(ApiException):
    """We got a response from the API but it didn't make any sense or we don't
    know how to handle it
    """
    pass

class ApiError(ApiException):
    """API gave us an error response (that we know how to parse)
    """
    pass

class ApiLoginFailure(ApiError):
    """Login info wasn't correct
    """
    pass

class ExpiredState(ApiException):
    """API state we tried to use is too old to be resumed from
    """
    pass



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import math
import hashlib
import zlib
import re
import logging

from tlslite.utils.cipherfactory import createAES

logger = logging.getLogger('crunchyroll.subtitles')

def aes_decrypt(key, iv, data):
    return str(createAES(key, iv).decrypt(data))

class SubtitleDecrypter(object):
    """Decrypt Crunchyroll's encrypted subtitle data

    This is pretty much copy and pasted from the Crunchyroll Takeout extension
    for XBMC, with minor readability enhancements

    @link https://code.google.com/p/urlxl-repo/source/browse/plugin.video.crunchyroll-takeout/resources/lib/crunchyDec.py
    """

    ENCRYPTION_KEY_SIZE     = 32
    # I have no idea where this comes from, gogo gadget copy and paste
    # results in 0x0540E9FA
    HASH_MAGIC_CONST        = int(math.floor(math.sqrt(6.9) * math.pow(2, 25)))

    HASH_SECRET_MOD_CONST   = 97
    HASH_SECRET_CHAR_OFFSET = 33
    HASH_SECRET_LENGTH      = 20

    def decrypt_subtitle(self, subtitle):
        """Decrypt encrypted subtitle data in high level model object

        @param crunchyroll.models.Subtitle subtitle
        @return str
        """
        return self.decrypt(self._build_encryption_key(int(subtitle.id)),
            subtitle['iv'][0].text.decode('base64'),
            subtitle['data'][0].text.decode('base64'))

    def decrypt(self, encryption_key, iv, encrypted_data):
        """Decrypt encrypted subtitle data

        @param int subtitle_id
        @param str iv
        @param str encrypted_data
        @return str
        """

        logger.info('Decrypting subtitles with length (%d bytes), key=%r',
            len(encrypted_data), encryption_key)
        return zlib.decompress(aes_decrypt(encryption_key, iv, encrypted_data))

    def _build_encryption_key(self, subtitle_id, key_size=ENCRYPTION_KEY_SIZE):
        """Generate the encryption key for a given media item

        Encryption key is basically just
        sha1(<magic value based on subtitle_id> + '"#$&).6CXzPHw=2N_+isZK') then
        padded with 0s to 32 chars

        @param int subtitle_id
        @param int key_size
        @return str
        """

        # generate a 160-bit SHA1 hash
        sha1_hash = hashlib.new('sha1', self._build_hash_secret((1, 2)) +
            self._build_hash_magic(subtitle_id)).digest()
        # pad to 256-bit hash for 32 byte key
        sha1_hash += '\x00' * max(key_size - len(sha1_hash), 0)
        return sha1_hash[:key_size]

    def _build_hash_magic(self, subtitle_id):
        """Build the other half of the encryption key hash

        I have no idea what is going on here

        @param int subtitle_id
        @return str
        """

        media_magic = self.HASH_MAGIC_CONST ^ subtitle_id
        hash_magic = media_magic ^ media_magic >> 3 ^ media_magic * 32
        return str(hash_magic)

    def _build_hash_secret(self, seq_seed, seq_len=HASH_SECRET_LENGTH,
            mod_value=HASH_SECRET_MOD_CONST):
        """Build a seed for the hash based on the Fibonacci sequence

        Take first `seq_len` + len(`seq_seed`) characters of Fibonacci
        sequence, starting with `seq_seed`, and applying e % `mod_value` +
        `HASH_SECRET_CHAR_OFFSET` to the resulting sequence, then return as
        a string

        @param tuple|list seq_seed
        @param int seq_len
        @param int mod_value
        @return str
        """

        # make sure we use a list, tuples are immutable
        fbn_seq = list(seq_seed)
        for i in range(seq_len):
            fbn_seq.append(fbn_seq[-1] + fbn_seq[-2])
        hash_secret = map(
            lambda c: chr(c % mod_value + self.HASH_SECRET_CHAR_OFFSET),
            fbn_seq[2:])
        return ''.join(hash_secret)

class SubtitleFormatter(object):
    """Base subtitle formatter class
    """

    def format(self, subtitles):
        """Turn a string containing the subs xml document into the formatted
        subtitle string

        @param str|crunchyroll.models.StyledSubtitle sub_xml_text
        @return str
        """
        logger.debug('Formatting subtitles (id=%s) with %s',
            subtitles.id, self.__class__.__name__)
        return self._format(subtitles).encode('utf-8')

    def _format(self, styled_subtitle):
        """Do the actual formatting on the parsed xml document, should be
        overridden by subclasses

        @param crunchyroll.models.StyledSubtitle styled_subtitle
        @return str
        """
        raise NotImplemented

class ASS4Formatter(SubtitleFormatter):
    """Subtitle formatter for ASS v4 format
    """

    STYLE_HEADER    = u'[V4 Styles]'
    STYLE_KEYS      = u'Format: Name, Fontname, Fontsize, PrimaryColour, ' \
        'SecondaryColour, OutlineColour, BackColour, Bold, Italic, Underline, ' \
        'StrikeOut, ScaleX, ScaleY, Spacing, Angle, BorderStyle, Outline, ' \
        'Shadow, Alignment, MarginL, MarginR, MarginV, Encoding'
    STYLE_FORMAT    = u'Style: {name}, {font_name}, {font_size}, {primary_colour}, ' \
        '{secondary_colour}, {outline_colour}, {back_colour}, {bold}, {italic}, ' \
        '{underline}, {strikeout}, {scale_x}, {scale_y}, {spacing}, {angle}, ' \
        '{border_style}, {outline}, {shadow}, {alignment}, {margin_l}, ' \
        '{margin_r}, {margin_v}, {encoding}'

    EVENT_HEADER    = u'[Events]'
    EVENT_KEYS      = u'Format: Layer, Start, End, Style, Name, MarginL, ' \
        'MarginR, MarginV, Effect, Text'
    EVENT_FORMAT    = u'Dialogue: 0,{start},{end},{style},{name},{margin_l},' \
        '{margin_r},{margin_v},{effect},{text}'

    def _format(self, styled_subtitle):
        return '\n'.join([
            self._format_header(styled_subtitle),
            self._format_styles(styled_subtitle.findall('.//styles/style')),
            self._format_events(styled_subtitle.findall('.//events/event')),
        ])

    def _format_header(self, subtitle_element):
        header = u"""[Script Info]
Title: {title}
ScriptType: v4.00
WrapStyle: {wrap_style}
PlayRexX: {play_res_x}
PlayResY: {play_res_y}
Subtitle ID: {id}
Language: {lang_string}
Created: {created}
"""
        return header.format(**subtitle_element._data.attrib)

    def _format_styles(self, style_elements):
        logger.debug('Formatting %d ASS style elements', len(style_elements))
        style = '\n'.join([
            self.STYLE_HEADER,
            self.STYLE_KEYS,
            '\n'.join(self._format_style(style) \
                for style in style_elements),
        ])
        return style + '\n'

    def _format_style(self, style_element):
        attrs = style_element._data.attrib.copy()
        for (k, v) in attrs.iteritems():
            # wikipedia suggests that v4 uses b10, while v4+ uses b16
            if v.startswith('&H'):
                attrs[k] = int(v[2:], 16)
        return self.STYLE_FORMAT.format(**attrs)

    def _format_events(self, event_elements):
        logger.debug('Formatting %d ASS event elements', len(event_elements))
        event_elements.sort(key=lambda e: int(e.id))
        events = '\n'.join([
            self.EVENT_HEADER,
            self.EVENT_KEYS,
            '\n'.join(self._format_event(event) \
                for event in event_elements),
        ])
        return events + '\n'

    def _format_event(self, event_element):
        return self.EVENT_FORMAT.format(**event_element._data.attrib)

class ASS4plusFormatter(ASS4Formatter):
    """Subtitle formatter for ASS v4+ format
    """

    STYLE_HEADER    = u'[V4+ Styles]'

    def _format_header(self, subtitle_element):
        header = u"""[Script Info]
Title: {title}
ScriptType: v4.00+
WrapStyle: {wrap_style}
PlayRexX: {play_res_x}
PlayResY: {play_res_y}
Subtitle ID: {id}
Language: {lang_string}
Created: {created}
"""
        return header.format(**subtitle_element._data.attrib)

    def _format_style(self, style_element):
        return self.STYLE_FORMAT.format(**style_element._data.attrib)

class SRTFormatter(SubtitleFormatter):
    """Subtitle formatter for SRT (unstyled) format
    """

    ASS_CMD_PATTERN     = re.compile(r'{[^}]+}')
    ASS_NEWLINE_PATTERN = re.compile(r'(?:\\n|\\N)')

    def _format(self, styled_subtitle):
        events = styled_subtitle.findall('.//events/event')
        logger.debug('Formatting %d SRT events', len(events))
        events.sort(key=lambda e: e.id)
        return '\n\n'.join(self._format_event(idx, event) \
            for idx, event in enumerate(events, 1)) + '\n'

    def _format_event(self, index, event):
        return '\n'.join([
            str(index),
            '{0} --> {1}'.format(
                self._format_timestamp(event.start),
                self._format_timestamp(event.end)),
            self._format_event_text(event),
        ])

    def _format_event_text(self, event):
        text = event._data.attrib.get('text')
        text = self.ASS_CMD_PATTERN.sub('', text)
        text = self.ASS_NEWLINE_PATTERN.sub('', text)
        return text

    def _format_timestamp(self, timestamp):
        """Format timestamp to what SRT wants

        Turns a string like 0:00:04.17 -> 00:00:04,170

        @param str timestamp
        @return str
        """
        return '{0:02d}:{1:02d}:{2:02d},{3:02d}0'.format(
            *map(int, timestamp.replace('.', ':').split(':')))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from setuptools import setup, find_packages

import crunchyroll

try:
    with open('README.md') as readme:
        long_description = readme.read()

    with open('requirements.txt') as reqs:
        requirements = [line.strip() for line in reqs if line.strip()]
except IOError:
    long_description = crunchyroll.__description__
    requirements = [
        'requests',
        'tlslite',
    ]

SETUP_ARGS = {
    # package metadata
    'name':             crunchyroll.__title__,
    'description':      crunchyroll.__description__,
    'long_description': long_description,
    'version':          crunchyroll.__version__,
    'author':           crunchyroll.__author__,
    'author_email':     crunchyroll.__author_email__,
    'url':              crunchyroll.__url__,

    # pypi metadata
    'license':          'GPLv2.0',
    'platforms':        'any',
    'install_requires': requirements,
    'classifiers':      [
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    # setuptools info
    'packages':         find_packages(),
    'test_suite':       'tests',
}

if __name__ == '__main__':
    setup(**SETUP_ARGS)



# -*- coding: utf-8 -*-

# Copyright (C) 2013  Alex Headley  <aheadley@waysaboutstuff.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import re
import json

from crunchyroll.apis import ApiInterface
from crunchyroll.constants import SCRAPER
from crunchyroll.apis.errors import *

class ScraperApi(ApiInterface):
    """Website scraper API
    """

    def __init__(self, connector):
        self._connector = connector

    def get_media_formats(self, media_id):
        """CR doesn't seem to provide the video_format and video_quality params
        through any of the APIs so we have to scrape the video page
        """
        url = (SCRAPER.API_URL + 'media-' + media_id).format(
            protocol=SCRAPER.PROTOCOL_INSECURE)
        format_pattern = re.compile(SCRAPER.VIDEO.FORMAT_PATTERN)
        formats = {}

        for format, param in SCRAPER.VIDEO.FORMAT_PARAMS.iteritems():
            resp = self._connector.get(url, params={param: '1'})
            if not resp.ok:
                continue
            match = format_pattern.search(resp.content)
            if match:
                formats[format] = (int(match.group(1)), int(match.group(2)))
        return formats
